from .dsp import ExpFilter

from .visualizer import Visualizer
from .scroll import Scroll
from .spectrum import Spectrum
from .energy import Energy

from .source import Source
from .microphone import Microphone

from .controller import Controller
from .sampler import Sampler



import numpy as np

from .dsp import ExpFilter

from .visualizer import Visualizer


class Spectrum(Visualizer):

    def __init__(self, num_pixels: int, **kwargs):
        self.num_pixels = num_pixels
        self.prev_spectrum = np.tile(0.01, self.num_pixels // 2)
        self.common_mode = ExpFilter(np.tile(0.01, self.num_pixels // 2),
                                     alpha_decay=0.99, alpha_rise=0.01)
        self.red_filter = ExpFilter(np.tile(0.01, self.num_pixels // 2),
                                    alpha_decay=0.2, alpha_rise=0.99)
        self.blue_filter = ExpFilter(np.tile(0.01, self.num_pixels // 2),
                                     alpha_decay=0.1, alpha_rise=0.5)

    def visualize(self, inp: np.ndarray) -> np.ndarray:
        y = np.copy(self._interpolate(inp, self.num_pixels // 2))
        self.common_mode.update(y)
        diff = y - self.prev_spectrum
        self.prev_spectrum = np.copy(y)
        # Color channel mappings
        r = self.red_filter.update(y - self.common_mode.value)
        g = np.abs(diff)
        b = self.blue_filter.update(np.copy(y))
        # Mirror the color channels for symmetric output
        return np.array([
            np.concatenate((r[::-1], r)),
            np.concatenate((g[::-1], g)),
            np.concatenate((b[::-1], b)),
        ]) * 255

    @staticmethod
    def _interpolate(y, new_length):
        if len(y) == new_length:
            return y
        return np.interp(
            np.linspace(0, 1, new_length),  # x_new
            np.linspace(0, 1, len(y)),  # x_old
            y,
        )



from numpy import ndarray


class ExpFilter:
    """Simple exponential smoothing filter"""

    def __init__(self, val=0.0, alpha_decay=0.5, alpha_rise=0.5):
        """Small rise / decay factors = more smoothing"""
        assert 0.0 < alpha_decay < 1.0, 'Invalid decay smoothing factor'
        assert 0.0 < alpha_rise < 1.0, 'Invalid rise smoothing factor'
        self.alpha_decay = alpha_decay
        self.alpha_rise = alpha_rise
        self.value = val

    def update(self, value):
        if isinstance(self.value, (list, ndarray, tuple)):
            alpha = value - self.value
            alpha[alpha > 0.0] = self.alpha_rise
            alpha[alpha <= 0.0] = self.alpha_decay
        else:
            alpha = self.alpha_rise if value > self.value else self.alpha_decay
        self.value = alpha * value + (1.0 - alpha) * self.value
        return self.value



import pyaudio
from numpy import ndarray, frombuffer, float32, int16

from .source import Source


class Microphone(Source):
    py_audio: pyaudio.PyAudio
    stream: pyaudio.Stream
    rate: int
    fps: int
    overflows: int = 0

    def __init__(self, rate: int = 44100, fps: int = 60):
        self.rate = rate
        self.fps = fps

    def __enter__(self):
        self.py_audio = pyaudio.PyAudio()
        self.stream = self.py_audio.open(format=pyaudio.paInt16,
                                         channels=1,
                                         rate=self.rate,
                                         input=True,
                                         frames_per_buffer=int(self.rate / self.fps))
        self.stream.start_stream()
        return self

    def audio_sample(self) -> ndarray:
        if not self.stream:
            return ndarray([])
        try:
            y = frombuffer(self.stream.read(int(self.rate / self.fps)), dtype=int16)
            y = y.astype(float32)
            # empty remaining buffer
            self.stream.read(self.stream.get_read_available())
            return y
        except IOError:
            self.overflows += 1
            print('Audio buffer has overflowed {} times'.format(self.overflows))
            return ndarray([])

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stream.stop_stream()
        self.stream.close()
        self.py_audio.terminate()



import numpy as np
from scipy.ndimage import gaussian_filter1d

from . import ExpFilter, Visualizer


class Energy(Visualizer):
    def __init__(self, num_pixels: int, num_frequency_bins: int = 24):
        self.num_pixels = num_pixels
        self.pixels = np.tile(1.0, (3, self.num_pixels // 2))
        self.gain = ExpFilter(np.tile(0.01, num_frequency_bins), alpha_decay=0.001, alpha_rise=0.99)
        self.pixel_filter = ExpFilter(np.tile(1, (3, self.num_pixels // 2)),
                                      alpha_decay=0.1, alpha_rise=0.99)

    def visualize(self, inp: np.ndarray) -> np.ndarray:
        y = np.copy(inp)
        self.gain.update(y)
        y /= self.gain.value
        # Scale by the width of the LED strip
        y *= float((self.num_pixels // 2) - 1)
        # Map color channels according to energy in the different freq bands
        scale = 0.9
        r = int(np.mean(y[:len(y) // 3] ** scale))
        g = int(np.mean(y[len(y) // 3: 2 * len(y) // 3] ** scale))
        b = int(np.mean(y[2 * len(y) // 3:] ** scale))
        # Assign color to different frequency regions
        self.pixels[0, :r] = 255.0
        self.pixels[0, r:] = 0.0
        self.pixels[1, :g] = 255.0
        self.pixels[1, g:] = 0.0
        self.pixels[2, :b] = 255.0
        self.pixels[2, b:] = 0.0
        self.pixel_filter.update(self.pixels)
        self.pixels = np.round(self.pixel_filter.value)
        # Apply substantial blur to smooth the edges
        self.pixels[0, :] = gaussian_filter1d(self.pixels[0, :], sigma=4.0)
        self.pixels[1, :] = gaussian_filter1d(self.pixels[1, :], sigma=4.0)
        self.pixels[2, :] = gaussian_filter1d(self.pixels[2, :], sigma=4.0)
        # Set the new pixel value
        return np.concatenate((self.pixels[:, ::-1], self.pixels), axis=1)



import numpy as np
from scipy.ndimage import gaussian_filter1d

from .dsp import ExpFilter
from .visualizer import Visualizer


class Scroll(Visualizer):
    def __init__(self, num_pixels: int, num_frequency_bins: int = 24):
        self.num_pixels = num_pixels
        self.pixels = np.tile(1.0, (3, num_pixels // 2))
        self.gain = ExpFilter(np.tile(0.01, num_frequency_bins), alpha_decay=0.001, alpha_rise=0.99)

    def visualize(self, inp: np.ndarray) -> np.ndarray:
        y = inp ** 2.0
        self.gain.update(y)
        y /= self.gain.value
        y *= 255.0
        r = int(np.max(y[:len(y) // 3]))
        g = int(np.max(y[len(y) // 3: 2 * len(y) // 3]))
        b = int(np.max(y[2 * len(y) // 3:]))
        # Scrolling effect window
        self.pixels[:, 1:] = self.pixels[:, :-1]
        self.pixels *= 0.98
        self.pixels = gaussian_filter1d(self.pixels, sigma=0.2)
        # Create new color originating at the center
        self.pixels[0, 0] = r
        self.pixels[1, 0] = g
        self.pixels[2, 0] = b
        # Update the LED strip
        return np.concatenate((self.pixels[:, ::-1], self.pixels), axis=1)



# Check https://github.com/scottlawsonbc/audio-reactive-led-strip/blob/master/python/melbank.py for original source.
from numpy import linspace, zeros, arange, log10


def hertz_to_mel(freq):
    """Returns mel-frequency from linear frequency input.
    Parameter
    ---------
    freq : scalar or ndarray
        Frequency value or array in Hz.
    Returns
    -------
    mel : scalar or ndarray
        Mel-frequency value or ndarray in Mel
    """
    return 2595.0 * log10(1 + (freq / 700.0))


def mel_to_hertz(mel):
    """Returns frequency from mel-frequency input.
    Parameter
    ---------
    mel : scalar or ndarray
        Mel-frequency value or ndarray in Mel
    Returns
    -------
    freq : scalar or ndarray
        Frequency value or array in Hz.
    """
    return 700.0 * (10 ** (mel / 2595.0)) - 700.0


def melfrequencies_mel_filterbank(num_bands, freq_min, freq_max):
    """Returns centerfrequencies and band edges for a mel filter bank
    Parameters
    ----------
    num_bands : int
        Number of mel bands.
    freq_min : scalar
        Minimum frequency for the first band.
    freq_max : scalar
        Maximum frequency for the last band.
    Returns
    -------
    center_frequencies_mel : ndarray
    lower_edges_mel : ndarray
    upper_edges_mel : ndarray
    """

    mel_max = hertz_to_mel(freq_max)
    mel_min = hertz_to_mel(freq_min)
    delta_mel = abs(mel_max - mel_min) / (num_bands + 1.0)
    frequencies_mel = mel_min + delta_mel * arange(0, num_bands + 2)
    lower_edges_mel = frequencies_mel[:-2]
    upper_edges_mel = frequencies_mel[2:]
    center_frequencies_mel = frequencies_mel[1:-1]
    return center_frequencies_mel, lower_edges_mel, upper_edges_mel


def compute_melmat(num_mel_bands=12, freq_min=64, freq_max=8000,
                   num_fft_bands=513, sample_rate=16000):
    """Returns tranformation matrix for mel spectrum.
    Parameters
    ----------
    num_mel_bands : int
        Number of mel bands. Number of rows in melmat.
        Default: 24
    freq_min : scalar
        Minimum frequency for the first band.
        Default: 64
    freq_max : scalar
        Maximum frequency for the last band.
        Default: 8000
    num_fft_bands : int
        Number of fft-frequenc bands. This ist NFFT/2+1 !
        number of columns in melmat.
        Default: 513   (this means NFFT=1024)
    sample_rate : scalar
        Sample rate for the signals that will be used.
        Default: 44100
    Returns
    -------
    melmat : ndarray
        Transformation matrix for the mel spectrum.
        Use this with fft spectra of num_fft_bands_bands length
        and multiply the spectrum with the melmat
        this will tranform your fft-spectrum
        to a mel-spectrum.
    frequencies : tuple (ndarray <num_mel_bands>, ndarray <num_fft_bands>)
        Center frequencies of the mel bands, center frequencies of fft spectrum.
    """
    center_frequencies_mel, lower_edges_mel, upper_edges_mel = \
        melfrequencies_mel_filterbank(
            num_mel_bands,
            freq_min,
            freq_max
        )

    center_frequencies_hz = mel_to_hertz(center_frequencies_mel)
    lower_edges_hz = mel_to_hertz(lower_edges_mel)
    upper_edges_hz = mel_to_hertz(upper_edges_mel)
    freqs = linspace(0.0, sample_rate / 2.0, num_fft_bands)
    melmat = zeros((num_mel_bands, num_fft_bands))

    for imelband, (center, lower, upper) in enumerate(zip(
            center_frequencies_hz, lower_edges_hz, upper_edges_hz)):
        left_slope = (freqs >= lower) == (freqs <= center)
        melmat[imelband, left_slope] = (
                (freqs[left_slope] - lower) / (center - lower)
        )

        right_slope = (freqs >= center) == (freqs <= upper)
        melmat[imelband, right_slope] = (
                (upper - freqs[right_slope]) / (upper - center)
        )

    return melmat, (center_frequencies_mel, freqs)



import socket
from time import time

from typing import Callable

from .dsp import ExpFilter
from .sampler import Sampler


class Controller:
    sampler: Sampler = None
    prev_time: float = time() * 1000.0

    def __init__(self, sampler: Sampler, on_sample: Callable = None, address: str = "127.0.0.1", port: int = 7777,
                 show_fps: bool = False):
        self.sampler = sampler
        self.on_sample = on_sample
        self.show_fps = show_fps
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._addr = address
        self._port = port
        self._fps = ExpFilter(val=sampler.source.fps, alpha_decay=0.2, alpha_rise=0.2)
        self._is_running = False

    def process_sample(self):
        self.send_data()
        self.on_sample(self.sampler.update_sample()) if self.on_sample else self.sampler.update_sample()

    def run(self):
        if self._is_running:
            return
        self._is_running = True
        while self._is_running:
            self.process_sample()

    def stop(self, *args):
        self._is_running = False

    def send_data(self):
        self._sock.sendto(self.sampler.sample(), (self._addr, self._port))

    def fps(self):
        time_now = time() * 1000.0
        dt = time_now - self.prev_time
        self.prev_time = time_now
        if dt == 0.0:
            return self._fps.value
        return self._fps.update(1000.0 / dt)



import numpy as np
from scipy.ndimage import gaussian_filter1d

from . import ExpFilter, Source, Visualizer
from .melbank import compute_melmat


class Sampler:
    y_rolling: np.ndarray
    source: Source
    _gamma_table = None

    def __init__(self, source: Source, visualizer: Visualizer, gamma_table_path: str = None, num_pixels: int = 60,
                 max_pixels_per_packet: int = 126, min_volume_threshold: int = 1e-7,
                 num_frames_rolling_window: int = 2, num_frequency_bins: int = 24,
                 min_freq: int = 200, max_freq: int = 12000
                 ):
        self.num_pixels = num_pixels
        self.source = source
        self.visualizer = visualizer
        self.pixels = np.tile(1, (3, self.num_pixels))
        self.prev_sample = np.tile(253, (3, self.num_pixels))
        self.y_rolling = np.random.rand(num_frames_rolling_window, int(source.rate / source.fps)) / 1e16
        self.fft_window = np.hamming(int(source.rate / source.fps) * num_frames_rolling_window)
        self.mel_gain = ExpFilter(np.tile(1e-1, num_frequency_bins),
                                  alpha_decay=0.01, alpha_rise=0.99)
        self.mel_smoothing = ExpFilter(np.tile(1e-1, num_frequency_bins),
                                       alpha_decay=0.5, alpha_rise=0.99)
        self.mel_y, _ = compute_melmat(num_mel_bands=num_frequency_bins,
                                       freq_min=min_freq,
                                       freq_max=max_freq,
                                       num_fft_bands=int(source.rate * num_frames_rolling_window / (2.0 * source.fps)),
                                       sample_rate=source.rate)
        self.min_vol = min_volume_threshold
        if gamma_table_path:
            self._gamma_table = np.load(gamma_table_path)
        if max_pixels_per_packet:
            self.max_pixels_per_packet = max_pixels_per_packet

    def sample(self) -> bytes:
        # Truncate values and cast to integer
        p = np.clip(self.pixels, 0, 255).astype(int)
        if self._gamma_table:
            p = self._gamma_table[p]
        idxs = [i for i in range(p.shape[1]) if not np.array_equal(p[:, i], self.prev_sample[:, i])]
        n_packets = len(idxs) // self.max_pixels_per_packet + 1
        idxs = np.array_split(idxs, n_packets)
        m = []
        for idx in idxs:
            for i in idx:
                m.append(i)  # Index of pixel to change
                m.append(p[0][i])  # Pixel red value
                m.append(p[1][i])  # Pixel green value
                m.append(p[2][i])  # Pixel blue value
        self.prev_sample = np.copy(p)
        return bytes(m)

    def update_sample(self) -> np.ndarray:
        y = self.source.audio_sample() / 2.0 ** 15
        self.y_rolling[:-1] = self.y_rolling[1:]
        self.y_rolling[-1, :] = np.copy(y)
        y_data = np.concatenate(self.y_rolling, axis=0).astype(np.float32)

        vol = np.max(np.abs(y_data))
        if vol < self.min_vol:
            self.pixels = np.tile(0, (3, self.num_pixels))
        else:
            rolling_len = len(y_data)
            n_zeros = 2 ** int(np.ceil(np.log2(rolling_len))) - rolling_len
            # Pad with zeros until the next power of two
            y_data *= self.fft_window
            y_padded = np.pad(y_data, (0, n_zeros), mode='constant')
            # Construct a Mel filterbank from the FFT data
            mel = np.atleast_2d(
                np.abs(np.fft.rfft(y_padded)[:rolling_len // 2])
            ).T * self.mel_y.T
            # Scale data to values more suitable for visualization
            mel = np.sum(mel, axis=0)
            mel = mel ** 2.0
            # Gain normalization
            self.mel_gain.update(np.max(gaussian_filter1d(mel, sigma=1.0)))
            mel /= self.mel_gain.value
            mel = self.mel_smoothing.update(mel)
            # Map filterbank output onto LED strip
            self.pixels = self.visualizer.visualize(mel)
            return self.pixels



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="neopixel",  # This is the name of the package
    version="0.0.1",  # The initial release version
    author="Ajat Prabha",  # Full name of the author
    description="An importable package to control LEDs via UDP protocol",
    license='MIT',
    url="https://github.com/ajatprabha/neopixel",
    platforms=['any'],
    long_description=long_description,  # Long description read from the the readme file
    long_description_content_type="text/markdown",
    packages=['neopixel'],  # List of all python modules to be installed
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],  # Information to filter the project on PyPi website
    python_requires='>=3.6',  # Minimum version requirement of the package
    py_modules=['neopixel'],  # Name of the python package
    package_dir={'neopixel': 'src'},  # Directory of the source code of the package
    install_requires=['numpy', 'scipy', 'pyaudio']  # Install other dependencies if any
)
